@@ -15,6 +15,8 @@
 
 #include_next <net/ip6_fib.h>
 
+#ifndef HAVE_RT6_GET_COOKIE
+
 #ifndef RTF_PCPU
 #define RTF_PCPU        0x40000000
 #endif
@@ -26,11 +28,16 @@
 #define rt6_get_cookie rpl_rt6_get_cookie
 static inline u32 rt6_get_cookie(const struct rt6_info *rt)
 {
-	if (rt->rt6i_flags & RTF_PCPU ||
-			(unlikely(rt->dst.flags & DST_NOCACHE) && rt->dst.from))
-		rt = (struct rt6_info *)(rt->dst.from);
+       if (rt->rt6i_flags & RTF_PCPU ||
+#ifdef HAVE_DST_NOCACHE
+           (unlikely(rt->dst.flags & DST_NOCACHE) && rt->dst.from))
+#else
+           (unlikely(!list_empty(&rt->rt6i_uncached)) && rt->dst.from))
+#endif
+               rt = (struct rt6_info *)(rt->dst.from);
 
-	return rt->rt6i_node ? rt->rt6i_node->fn_sernum : 0;
+       return rt->rt6i_node ? rt->rt6i_node->fn_sernum : 0;
 }
+#endif
 
 #endif
