@@ -48,6 +48,8 @@ struct bond_settings {
     int rebalance_interval;     /* Milliseconds between rebalances.
                                    Zero to disable rebalancing. */
 
+    const char *primary;        /* For AB mode, primary interface name. */
+
     /* Link status detection. */
     int up_delay;               /* ms before enabling an up slave. */
     int down_delay;             /* ms before disabling a down slave. */
@@ -58,6 +60,8 @@ struct bond_settings {
                                 /* The MAC address of the interface
                                    that was active during the last
                                    ovs run. */
+    bool use_lb_output_action;  /* Use lb_output action. Only applicable for
+                                   bond mode BALANCE TCP. */
 };
 
 /* Program startup. */
@@ -122,6 +126,7 @@ void bond_rebalance(struct bond *);
 */
 void bond_update_post_recirc_rules(struct bond *, uint32_t *recirc_id,
                                    uint32_t *hash_basis);
-bool bond_may_recirc(const struct bond *, uint32_t *recirc_id,
-                     uint32_t *hash_bias);
+
+bool bond_use_lb_output_action(const struct bond *bond);
+
 #endif /* bond.h */
